/*****************************************************************************
*
* @brief Startup file.
*
* @author evgeny.linsky@cloudbear.ru, vitaly.gaiduk@cloudbear.ru
*
* Copyright (c) 2018 CloudBEAR LLC - http://www.cloudbear.ru/
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   Redistributions of source code must retain the above copyright
*   notice, this list of conditions and the following disclaimer.
*
*   Redistributions in binary form must reproduce the above copyright
*   notice, this list of conditions and the following disclaimer in the
*   documentation and/or other materials provided with the
*   distribution.
*
*   Neither the name of CloudBEAR nor the names of
*   its contributors may be used to endorse or promote products derived
*   from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
******************************************************************************/

//#include "bear_uart.h"

#define MSTATUS_FS   0x00006000

#if __riscv_xlen == 64
# define LREG ld
# define SREG sd
# define REGBYTES 8
#else
# define LREG lw
# define SREG sw
# define REGBYTES 4
#endif

  .section ".text.init"
  .globl _start

_start:
  // Only first hart runs
  csrr s10, mhartid
  beqz s10, 1f
  j .
1:

  // Zeroing registers
  li  x1, 0
  li  x2, 0
  li  x3, 0
  li  x4, 0
  li  x5, 0
  li  x6, 0
  li  x7, 0
  li  x8, 0
  li  x9, 0
  li  x10,0
  li  x11,0
  li  x12,0
  li  x13,0
  li  x14,0
  li  x15,0
  li  x16,0
  li  x17,0
  li  x18,0
  li  x19,0
  li  x20,0
  li  x21,0
  li  x22,0
  li  x23,0
  li  x24,0
  li  x25,0
  li  x26,0
  li  x27,0
  li  x28,0
  li  x29,0
  li  x30,0
  li  x31,0

  // Global pointer (gp) allows using more effective gp-relative addressing instead of
  // abosulte/pc-relative for arbitrary 4KiB memory range. For the optimisation to be effective,
  // it should cover the most intensely used RAM area. For standard newlib apps, this is the
  // area where the .sdata section is located. The linker uses the (!) __global_pointer$ symbol
  // definition to compare the memory addresses and, if within range, it replaces
  // absolute/pc-relative addressing with gp-relative addressing, which makes the code more
  // efficient. This process is also called relaxing, and can be disabled by -Wl,--no-relax.
  // The gp register should be loaded during startup and should not be changed later.
  .option push
  .option norelax
    la gp, __global_pointer$
  .option pop

/* Calculate and set stack ponter (sp) for the one hart.                                    *
 * Also calculate base address and size then store them to brk_base and brk_size variables. *
 * Calculations are made according to the memory layout and equations presented below.      *
 *                                                                                          *
 *                                                                                          *
 *__vma_text_start +-------------------------+                                              *
 *                 |                         |                                              *
 *                 |  .text                  |                                              *
 *                 +-------------------------+                                              *
 *                                                                                          *
 * __vma_data_start+-------------------------+                                              *
 *                 |  .data                  |                                              *
 *                 |  .bss                   |                                              *
 *                 |  etc                    |                                              *
 *                 |                         |                                              *
 * __vma_brk_start +-------------------------+                                              *
 *                 |                         |                                              *
 *                 |  Heap area              |                                              *
 *                 |                         |                                              *
 *        brk_end  +------------------------------------------------+                       *
 *                 |                         |   |                  |                       *
 *                 |  Stack area             |   |   Hart 0 stack   |                       *
 *                 |                         |   |                  |                       *
 * tcm_end         +-------------------------+   +------------------+                       *
 *                                                                                          *
 *                                                                                          *
 *                                                                                          *
 *                                                                                          *
 *                                                                                          *
 *     brk_base =  end of data section                                                      *
 *     stack_size = 2 KiB                                                                   *
 *     brk_end = tcm_end - stack_size                                                       *
 *     brk_size = brk_end - __vma_brk_start                                                 *
 *                                                                                          *
 *     sp = __vma_end                                                                       *
 *                                                                                          *
 *     Note: variables starting with underscore are set by linker script.                   */

  /*----------------------------------*/
  /* Calculating heap and stack areas */
  /*----------------------------------*/

  la   sp, TCM_END_ADDR

  // Calculate brk_end and store it to s6
  // brk_end = __vma_end(sp) - stack_size
  // Stack size = 2 KiB.
  li s4, 2048
  sub s6, sp, s4

  // copy data section from LMA to VMA
  lui s2, %hi(__lma_data_start)
  addi s2, s2, %lo(__lma_data_start)

  lui s3, %hi(__vma_data_start)
  addi s3, s3, %lo(__vma_data_start)

  lui s5, %hi(__vma_data_end)
  addi s5, s5, %lo(__vma_data_end)

1:
  LREG t0, (s2)
  SREG t0, (s3)
  addi s2, s2, 4
  addi s3, s3, 4
  bltu s3, s5, 1b

  // brk_base = __end
  la s4, __vma_brk_start
  la t0, brk_base
  SREG s4, 0(t0)

  // brk_size = brk_end (s6) - brk_base (s4)
  sub s6, s6, s4
  la t0, brk_size
  SREG s6, 0(t0)

  /*----------*/
  /* Init FPU */
  /*----------*/

#ifdef __riscv_flen
  // Initialize mstatus and fcsr regs when fp is used.
  li t0, MSTATUS_FS & (MSTATUS_FS >> 1)
  csrw mstatus, t0
  csrwi fcsr, 0
#else
  li t0, 0
  csrw mstatus, t0
#endif // __riscv_flen

  /*-----------*/
  /* Init UART */
  /*-----------*/
  // Init UART early, before init trap, because trap handler
  // prints exception info via UART.

  //li a0, BEAR_UART_REF_CLK_MHZ
  //li a1, BEAR_UART_BAUD_RATE
  //jal uart_init

  /*-----------*/
  /* Init trap */
  /*-----------*/

  la t0, trap_vector
  csrw mtvec, t0

  /*-------------*/
  /* Zeroing bss */
  /*-------------*/

  la    t0, __bss_start
  la    t1, __bss_end
1:
  bge   t0, t1, 2f
  SREG  zero, 0(t0)
  addi  t0, t0, REGBYTES
  j     1b
2:

  /*--------------------------------------*/
  /* Continue initialization in user mode */
  /*--------------------------------------*/

  //la t0, _bear_init
  //csrw mepc, t0
  //mret // Switch to U-mode
  call main
  /*-------------*/
  /* Trap vector */
  /*-------------*/
  // Trap vector must be aligned on a 4-byte boundary
.align 2
trap_vector:
  addi sp, sp, -32*REGBYTES

  SREG x1, 1*REGBYTES(sp)
  SREG x2, 2*REGBYTES(sp)
  SREG x3, 3*REGBYTES(sp)
  SREG x4, 4*REGBYTES(sp)
  SREG x5, 5*REGBYTES(sp)
  SREG x6, 6*REGBYTES(sp)
  SREG x7, 7*REGBYTES(sp)
  SREG x8, 8*REGBYTES(sp)
  SREG x9, 9*REGBYTES(sp)
  SREG x10, 10*REGBYTES(sp)
  SREG x11, 11*REGBYTES(sp)
  SREG x12, 12*REGBYTES(sp)
  SREG x13, 13*REGBYTES(sp)
  SREG x14, 14*REGBYTES(sp)
  SREG x15, 15*REGBYTES(sp)
  SREG x16, 16*REGBYTES(sp)
  SREG x17, 17*REGBYTES(sp)
  SREG x18, 18*REGBYTES(sp)
  SREG x19, 19*REGBYTES(sp)
  SREG x20, 20*REGBYTES(sp)
  SREG x21, 21*REGBYTES(sp)
  SREG x22, 22*REGBYTES(sp)
  SREG x23, 23*REGBYTES(sp)
  SREG x24, 24*REGBYTES(sp)
  SREG x25, 25*REGBYTES(sp)
  SREG x26, 26*REGBYTES(sp)
  SREG x27, 27*REGBYTES(sp)
  SREG x28, 28*REGBYTES(sp)
  SREG x29, 29*REGBYTES(sp)
  SREG x30, 30*REGBYTES(sp)
  SREG x31, 31*REGBYTES(sp)

  csrr a0, mcause
  csrr a1, mepc
  mv a2, sp
  jal bear_gdb_exit
  csrw mepc, a0

  LREG x1, 1*REGBYTES(sp)
  LREG x2, 2*REGBYTES(sp)
  LREG x3, 3*REGBYTES(sp)
  LREG x4, 4*REGBYTES(sp)
  LREG x5, 5*REGBYTES(sp)
  LREG x6, 6*REGBYTES(sp)
  LREG x7, 7*REGBYTES(sp)
  LREG x8, 8*REGBYTES(sp)
  LREG x9, 9*REGBYTES(sp)
  LREG x10, 10*REGBYTES(sp)
  LREG x11, 11*REGBYTES(sp)
  LREG x12, 12*REGBYTES(sp)
  LREG x13, 13*REGBYTES(sp)
  LREG x14, 14*REGBYTES(sp)
  LREG x15, 15*REGBYTES(sp)
  LREG x16, 16*REGBYTES(sp)
  LREG x17, 17*REGBYTES(sp)
  LREG x18, 18*REGBYTES(sp)
  LREG x19, 19*REGBYTES(sp)
  LREG x20, 20*REGBYTES(sp)
  LREG x21, 21*REGBYTES(sp)
  LREG x22, 22*REGBYTES(sp)
  LREG x23, 23*REGBYTES(sp)
  LREG x24, 24*REGBYTES(sp)
  LREG x25, 25*REGBYTES(sp)
  LREG x26, 26*REGBYTES(sp)
  LREG x27, 27*REGBYTES(sp)
  LREG x28, 28*REGBYTES(sp)
  LREG x29, 29*REGBYTES(sp)
  LREG x30, 30*REGBYTES(sp)
  LREG x31, 31*REGBYTES(sp)

  addi sp, sp, 32*REGBYTES
  mret

  /*-------------------*/
  /* Exit and ret code */
  /*-------------------*/
// Code below is used for testing in CloudBEAR
// You can delete this or change for something yours

  .globl transform_code
transform_code:
// We should transform ret code this way
// code <<= 1, code |= 1;
  slli t0, t0, 1
  addi t0, t0, 1

  .globl bear_gdb_exit
bear_gdb_exit:
  j .

  .global _init
  .global _fini
_init:
_fini:
  // These don't have to do anything since we use init_array/fini_array.
  ret

.data

.align 3
.globl brk_base
brk_base:
  .dword 0x0

.globl brk_size
brk_size:
  .dword 0x0

